/*
 * Copyright (C) 2022 EXOLIGENT (www.exoligent.com)
 * Author: Maxime Coroyer <maxime.coroyer@exoligent.com>
 */

#define _GNU_SOURCE     // <sched.h>
#include <sched.h>      // CPU_ZERO(), CPU_SET()
#include <pthread.h>    // pthread()

#include <open62541/plugin/pubsub_ethernet.h>   // UA_EthernetWriterGroupTransportDataType

#include <linux/if_link.h>
#include <linux/if_xdp.h>

//--!! TODO: should not be used
#include "open62541/ua_pubsub.h"                // UA_WriterGroup

#include "opc-rt.h"
#include "gw.h"

#define US_TO_NS (1000)
#define MS_TO_NS (1000 * (US_TO_NS))
#define S_TO_NS (1000 * (MS_TO_NS))
#define SLEEP_S 5
#define CLOCKID CLOCK_TAI

static pthread_t thread_creation(UA_Int16 thr_prio, size_t core_affinity,
        void *(*thread) (void *), char *app_name, void *srv_cfg);
static void *pubsub_task(void *arg);

static struct timespec basetime;

/**
 * Cyclic callback (add, remove, change period)
 */
UA_StatusCode opc_rt_cb_add(UA_Server *srv, UA_NodeId group_id,
    UA_ServerCallback cb, void *data, UA_Double interval_ms,
    UA_DateTime *baseTime, UA_TimerPolicy timerPolicy, UA_UInt64 *cb_id)
{

    if (UA_NodeId_equal(&group_id, &gw.opc.pub.w_group_id)) {
        gw.opc.pubsub_thr_args.pub_data       = data;
        gw.opc.pubsub_thr_args.pub_callback   = cb;
        gw.opc.pubsub_thr_args.interval_ms    = interval_ms;
    } else {
        gw.opc.pubsub_thr_args.sub_data       = data;
        gw.opc.pubsub_thr_args.sub_callback   = cb;
    }

    return UA_STATUSCODE_GOOD;
}

UA_StatusCode opc_rt_cb_change(UA_Server *srv, UA_NodeId group_id,
    UA_UInt64 cb_id, UA_Double interval_ms, UA_DateTime *baseTime,
    UA_TimerPolicy timerPolicy)
{
    /**
     * Callback interval need not be modified as it is thread based
     * implementation. The thread uses nanosleep for calculating cycle time
     * and modification in nanosleep value changes cycle time
     */
    return UA_STATUSCODE_GOOD;
}

void opc_rt_cb_remove(UA_Server *srv, UA_NodeId group_id, UA_UInt64 cb_id)
{
    if (cb_id && (pthread_join(cb_id, NULL) != 0))
        _print(opc, warn, "pthread_join failed (thread_id=%"PRIu64")\n", cb_id);
}

/**
 * If the external data source is written over the information model, the
 * externalDataWriteCallback will be triggered. The user has to take care and assure
 * that the write leads not to synchronization issues and race conditions.
 */
UA_StatusCode opc_rt_cb_ext_datawrite(UA_Server *srv,
    const UA_NodeId *sess_id, void *sess_ctx, const UA_NodeId *node_id,
    void *node_ctx, const UA_NumericRange *range,
    const UA_DataValue *data_value)
{
    // node values are updated by using variables in the memory
    // UA_Server_write is not used for updating node values.
    return UA_STATUSCODE_GOOD;
}

UA_StatusCode opc_rt_cb_ext_dataread_notif(UA_Server *srv,
    const UA_NodeId *sess_id, void *sess_ctx, const UA_NodeId *node_id,
    void *node_ctx, const UA_NumericRange *range)
{
    // allow read without any preparation
    return UA_STATUSCODE_GOOD;
}

int opc_rt_app_start(void *args)
{
    struct gateway *gw = (struct gateway *) args;
    char thr_name[11] = "pubsub_app";

    gw->opc.pubsub_thr_task = thread_creation(OPC_RT_PUBSUB_SCHED_PRIORITY,
        OPC_RT_PUBSUB_CORE, pubsub_task, thr_name, args);
    return 0;
}

int opc_rt_app_stop(void *args)
{
    struct gateway *gw = (struct gateway *) args;
    pthread_join(gw->opc.pubsub_thr_task, NULL);
    return 0;
}

static pthread_t thread_creation(UA_Int16 thr_prio, size_t core_affinity,
        void *(*thread) (void *), char *task_name, void *srv_cfg)
{
    UA_Int32 ret = 0;
    struct sched_param thr_param;
    pthread_t thr_id = pthread_self();
    cpu_set_t cpu_set;

    thr_param.sched_priority = thr_prio;
    if (pthread_setschedparam(thr_id, SCHED_FIFO, &thr_param))
    {
        _print(opc, err, "%s task: pthread_setschedparam failed\n", task_name);
        exit(1);
    }

    _print(opc, info, "%s task: thread priority is %d\n", task_name,
        thr_param.sched_priority);

    CPU_ZERO(&cpu_set);
    CPU_SET(core_affinity, &cpu_set);

    if ((ret = pthread_setaffinity_np(thr_id, sizeof(cpu_set_t), &cpu_set)))
    {
        fprintf(stderr, "%s: pthread_setaffinity_np error (%s)",
            task_name, strerror(ret));
        exit(1);
    }

    if (pthread_create(&thr_id, NULL, thread, srv_cfg))
        _print(opc, warn, "%s task: cannot create thread\n", task_name);

    if (CPU_ISSET(core_affinity, &cpu_set))
        _print(opc, info, "%s task: cpu_core=%ld\n", task_name, core_affinity);

    return thr_id;
}

/* nanosecond field in timespec is checked for overflowing */
static void ns_check_n_reformat(struct timespec *value)
{
    while (value->tv_nsec > (S_TO_NS - 1)) {
        value->tv_sec  += 1;
        value->tv_nsec -= S_TO_NS;
    }
}

static void usr_app(struct gateway *gw)
{
#if defined FIP_RW_OP_VIA_RT_APP_TASK
    if (gw->fip.rw_op) {
        UA_LOCK(&gw->opc.mu);
        if (fip_r_op(gw))
            running = 0;
        if (fip_w_op(gw))
            running = 0;
        gw->fip.rw_op = 0;
        UA_UNLOCK(&gw->opc.mu);
    }
    else if (gw->fip.wr_op) {
        UA_LOCK(&gw->opc.mu);
        if (fip_w_op(gw))
            running = 0;
        if (fip_r_op(gw))
            running = 0;
        gw->fip.wr_op = 0;
        UA_UNLOCK(&gw->opc.mu);
    }
    else if (gw->fip.r_op) {
        UA_LOCK(&gw->opc.mu);
        if (fip_r_op(gw))
            running = 0;
        gw->fip.r_op = 0;
        UA_UNLOCK(&gw->opc.mu);
    }
    else if (gw->fip.w_op) {
        UA_LOCK(&gw->opc.mu);
        if (fip_w_op(gw))
            running = 0;
        gw->fip.w_op = 0;
        UA_UNLOCK(&gw->opc.mu);
    }
#endif
}

/**
 * Publisher/Subscriber thread routine
 */
static void *pubsub_task(void *arg)
{
    struct gateway *gw = (struct gateway *) arg;
    struct timespec next_ns_time = {.tv_sec = 0, .tv_nsec = 0};
    UA_ServerCallback pub_cb, sub_cb;
    UA_Server *srv = NULL;
    UA_UInt64 pub_sleep_ns, pub_alive_ns, interval_ns, tx_time;
    UA_WriterGroup *w_group = NULL; //--!! TODO: remove w_group usage
    UA_ReaderGroup *r_group = NULL; //--!! TODO: remove r_group usage
    UA_EthernetWriterGroupTransportDataType eth_transp_settings;
    UA_ExtensionObject transp_settings;

    /**
     * Publisher will sleep for 80% of cycle time and then prepares
     * the tx packet within 20%
     */
    pub_sleep_ns = gw->opc.pub.publish_period_us * US_TO_NS * OPC_RT_PUB_WKUP_PERCENTAGE;
    pub_alive_ns = (gw->opc.pub.publish_period_us * US_TO_NS) - pub_sleep_ns;

    srv         = gw->opc.srv;
    pub_cb      = gw->opc.pubsub_thr_args.pub_callback;
    sub_cb      = gw->opc.pubsub_thr_args.sub_callback;
    w_group     = (UA_WriterGroup *) gw->opc.pubsub_thr_args.pub_data;
    r_group     = (UA_ReaderGroup *) gw->opc.pubsub_thr_args.sub_data;
    interval_ns = (UA_UInt64) (gw->opc.pubsub_thr_args.interval_ms * MS_TO_NS);

    memset(&eth_transp_settings, 0, sizeof(UA_EthernetWriterGroupTransportDataType));
    memset(&transp_settings, 0, sizeof(UA_ExtensionObject));

    eth_transp_settings.transmission_time = 0;
    transp_settings.content.decoded.data = &eth_transp_settings;
    w_group->config.transportSettings = transp_settings;

    /* get current time and compute the next nanosleeptime */
    clock_gettime(CLOCKID, &basetime);
    next_ns_time.tv_sec  = basetime.tv_sec;
    next_ns_time.tv_nsec = basetime.tv_nsec;
    ns_check_n_reformat(&next_ns_time);

    while (gw->opc.running) {
        clock_nanosleep(CLOCKID, TIMER_ABSTIME, &next_ns_time, NULL);

        UA_LOCK(&gw->opc.mu);
        gw->opc.task_in_progress = 1;

        /* update next w_group tx time */
        tx_time = ((UA_UInt64) next_ns_time.tv_sec * S_TO_NS + \
                   (UA_UInt64) next_ns_time.tv_nsec) + \
                   pub_alive_ns + OPC_RT_QBV_OFFSET;
        eth_transp_settings.transmission_time = tx_time;

        //_print(opc, info, "pubsub_task - pubsub_cb\n");
        sub_cb(srv, r_group);
        gw->stats.nb_sub_op++;

        usr_app(gw);

        pub_cb(srv, w_group);
        gw->stats.nb_pub_op++;

        next_ns_time.tv_nsec += (__syscall_slong_t) interval_ns;
        ns_check_n_reformat(&next_ns_time);

        gw->opc.task_in_progress = 0;
        UA_UNLOCK(&gw->opc.mu);
    }

    return (void*) NULL;
}